/**
 *   @file  salldosal.c
 *
 *   @brief   
 *      This is the OS abstraction layer and is used by the the SA LLD.
 *
 *  \par
 *  ============================================================================
 *  @n   (C) Copyright 2009, Texas Instruments, Inc.
 * 
 *  Redistribution and use in source and binary forms, with or without 
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the   
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
*/
#include "../unittest.h"
#include <ti/drv/sa/sa_osal.h>

/* CSL CHIP, SEM Functional layer includes */
#include <ti/csl/csl_chip.h>
#include <ti/csl/csl_semAux.h>


/**********************************************************************
 ****************************** Defines *******************************
 **********************************************************************/

#define SALLD_HW_SEM     1

/**********************************************************************
 ************************** Global Variables **************************
 **********************************************************************/
static UInt32 salldMallocCounter = 0;
static UInt32 salldFreeCounter   = 0;

/**********************************************************************
 ******************************* Macros *******************************
 **********************************************************************/

/**********************************************************************
 *************************** OSAL Functions **************************
 **********************************************************************/
/**
 *  @b Description
 *  @n  
 *      The function is used to allocate a memory block of the specified size.
 *
 *      Note: If the LLD is used by applications on multiple core, the "salldHeap"
 *      should be in shared memory
 *
 *  @param[in]  num_bytes
 *      Number of bytes to be allocated.
 *
 *  @retval
 *      Allocated block address
 */
Ptr salld_sim_malloc (UInt32 num_bytes, int alignmnet)
{
	Error_Block	errorBlock;

    /* Increment the allocation counter. */
    salldMallocCounter++;	

	/* Allocate memory. */
    return (Memory_alloc (NULL, num_bytes, alignmnet, &errorBlock));
}

/**
 *  @b Description
 *  @n  
 *      The function is used to free a memory block of the specified size allocated 
 *      using Osal_saMalloc() API.
 *
 *  @param[in]  ptr
 *      Pointer to the memory block to be cleaned up.
 *
 *  @param[in]  size
 *      Size of the memory block to be cleaned up.
 *
 *  @retval
 *      Not Applicable
 */
Void salld_sim_free (Ptr ptr, UInt32 size)
{
    /* Increment the free counter. */
    salldFreeCounter++;	
    Memory_free (NULL, ptr, size);
}

/**
 * @brief   The macro is used by the SA LLD to indicate that a block
 * of memory is about to be accessed. If the memory block is cached then
 * this indicates that the application would need to ensure that the cache
 * is updated with the data from the actual memory.
 *
 * <b> Prototype: </b>
 *  The following is the C prototype for the expected OSAL API.
 *
 *  @verbatim
        void Osal_saBeginMemAccess (void* addr, uint32_t sizeWords)
    @endverbatim
 *
 *  <b> Parameters </b>
 *  @n  Address of memory block.
 *  @n  Size of memory block.
 */
Void Osal_saBeginMemAccess (void* addr, UInt32 size)
{
    #ifndef ARM11
    
    uint32_t    key;

    /* Disable Interrupts */
    key = Hwi_disable();

    /* Cleanup the prefetch buffer also. */
    CSL_XMC_invalidatePrefetchBuffer();

    SYS_CACHE_INV (addr, size, CACHE_FENCE_WAIT);

    asm   (" nop      4");
    asm   (" nop      4");
    asm   (" nop      4");
    asm   (" nop      4");
    
    /* Reenable Interrupts. */
    Hwi_restore(key);

    #endif     
        
}
/**
 * @brief   The macro is used by the SA LLD to indicate that the block of 
 * memory has finished being updated. If the memory block is cached then the 
 * application would need to ensure that the contents of the cache are updated
 * immediately to the actual memory.
 *
 * <b> Prototype: </b>
 *  The following is the C prototype for the expected OSAL API.
 *
 *  @verbatim
        void Osal_saEndMemAccess (void* addr, uint32_t sizeWords)
    @endverbatim
 *
 *  <b> Parameters </b>
 *  @n  Address of memory block.
 *  @n  Size of memory block.
 */


Void Osal_saEndMemAccess   (void* addr, UInt32 size)
{
    
    #ifndef ARM11
    uint32_t    key;

    /* Disable Interrupts */
    key = Hwi_disable();

    SYS_CACHE_WB (addr, size, CACHE_FENCE_WAIT);

    asm   (" nop      4");
    asm   (" nop      4");
    asm   (" nop      4");
    asm   (" nop      4");
    
    /* Reenable Interrupts. */
    Hwi_restore(key);
    #endif     
    
}

/**
 * @brief   The macro is used by the SA LLD to indicate that the security
 * context byuffer is about to be accessed. If the security context buffer 
 * is cacheable then this indicates that the application would need to ensure 
 * that the cache is updated with the data from the actual memory since the
 * security context will be updated by SASS Cache engine.
 * If the security context buffers are non-cacheable then these macros can 
 * be defined to be NOP. 
 *
 * <b> Prototype: </b>
 *  The following is the C prototype for the expected OSAL API.
 *
 *  @verbatim
        void Osal_saBeginScAccess (void* addr, uint32_t sizeWords)
    @endverbatim
 *
 *  <b> Parameters </b>
 *  @n  Address of memory block.
 *  @n  Size of memory block.
 */

Void Osal_saBeginScAccess (void* addr, UInt32 size)
{
    #ifndef ARM11
    
    uint32_t    key;

    /* Disable Interrupts */
    key = Hwi_disable();

    /* Cleanup the prefetch buffer also. */
    CSL_XMC_invalidatePrefetchBuffer();

    SYS_CACHE_INV (addr, size, CACHE_FENCE_WAIT);

    asm   (" nop      4");
    asm   (" nop      4");
    asm   (" nop      4");
    asm   (" nop      4");
    
    /* Reenable Interrupts. */
    Hwi_restore(key);

    #endif     
        
}



/**
 * @brief   The macro is used by the SA LLD to indicate that the security 
 * context buffer has finished being updated. If the memory block is cacheable 
 * then the application would need to ensure that the contents of the cache are 
 * updated immediately to the actual memory.
 * If the security context buffers are non-cacheable then these macros can 
 * be defined to be NOP. 
 *
 * <b> Prototype: </b>
 *  The following is the C prototype for the expected OSAL API.
 *
 *  @verbatim
        void Osal_saEndScAccess (void* addr, uint32_t sizeWords)
    @endverbatim
 *
 *  <b> Parameters </b>
 *  @n  Address of memory block.
 *  @n  Size of memory block.
 */
 
Void Osal_saEndScAccess   (void* addr, UInt32 size)
{
    #ifndef ARM11
    uint32_t    key;

    /* Disable Interrupts */
    key = Hwi_disable();

    SYS_CACHE_WB (addr, size, CACHE_FENCE_WAIT);

    asm   (" nop      4");
    asm   (" nop      4");
    asm   (" nop      4");
    asm   (" nop      4");
    
    /* Reenable Interrupts. */
    Hwi_restore(key);
    #endif     
    
}

#if 0

/**
 *  @b Description
 *  @n  
 *      The function is used to enter a critical section.
 *      Function protects against 
 *      access from multiple threads on single core
 *
 *  @param[in]  key
 *      Key used to lock the critical section.
 *
 *  @retval
 *      Not Applicable
 */
Void Osal_saCsEnter (UInt32 *key)
{
#ifdef USE_BIOS
  *key = (UInt32)GateHwi_enter(tFramework.gateHwi);	
#else
  #ifndef ARM11
  *key = (UInt32)thwCriticalBeginFast();
  #else
  *key = 0x5000;
  #endif
#endif  

}

/**
 *  @b Description
 *  @n  
 *      The function is used to exit a critical section 
 *      protected using Osal_salldCsEnter() API.
 *
 *  @param[in]  key
 *      Key used to unlock the critical section.
 *
 *  @retval
 *      Not Applicable
 */
Void Osal_saCsExit (UInt32 key)
{
#ifdef USE_BIOS
  GateHwi_leave(tFramework.gateHwi, (UInt)key);
#else
  #ifndef ARM11 
  thwCriticalEndFast(key);
  #endif
#endif  

}

#endif

/**
 *  @b Description
 *  @n  
 *      The function is used to enter a critical section.
 *      Function protects against 
 *
 *      access from multiple threads on single core
 *      and
 *      access from multiple cores 
 *
 *  @param[in]  key
 *      Key used to lock the critical section.
 *
 *  @retval
 *      Not Applicable
 */
Void Osal_saMtCsEnter (UInt32 *key)
{
#ifdef USE_BIOS
	Semaphore_pend (tFramework.tfSaSem, BIOS_WAIT_FOREVER);
    *key = (UInt32)tFramework.tfSaSem;
#else
    #ifndef ARM11
    *key = (UInt32)thwCriticalBeginFast();
    #else
    *key = 0x6000;
    #endif
#endif    

}

/**
 *  @b Description
 *  @n  
 *      The function is used to exit a critical section 
 *      protected using Osal_salldCsEnter() API.
 *
 *  @param[in]  key
 *      Key used to unlock the critical section.
 *
 *  @retval
 *      Not Applicable
 */
Void Osal_saMtCsExit (UInt32 key)
{
#ifdef USE_BIOS
  Semaphore_post ((Semaphore_Handle)key);
#else
  #ifndef ARM11 
  thwCriticalEndFast(key);
  #endif
#endif  

}

/**
 *  @b Description
 *  @n  
 *      The function is the SA LLD OSAL Logging API which logs 
 *      the messages on the console.
 *
 *  @param[in]  fmt
 *      Formatted String.
 *
 *  @retval
 *      Not Applicable
 */
Void Osal_saLog ( String fmt, ... )
{

}

void* Osal_saGetSCPhyAddr(void* vaddr)
{
    return vaddr;
}

uint16_t Osal_saGetProcId (void )
{
    return (uint16_t)CSL_chipReadReg(CSL_CHIP_DNUM);
}


/**
 * @brief   The macro is used by the SA LLD to the Endian mode of the system (SoC).
 *
 * <b> Prototype: </b>
 *  The following is the C prototype for the expected OSAL API.
 *
 *  @verbatim
 *      int Osal_saGetSysEndianMode(void) 
 *   @endverbatim
 *
 *  <b> Return Value Endian mode of the system (SoC) </b>
 *  <b> Parameters </b>
 *  @n  Endian mode of the system (SoC).
 */ 
int   Osal_saGetSysEndianMode(void)
{
#if defined( _BIG_ENDIAN ) 
    return((int)sa_SYS_ENDIAN_MODE_BIG);
#else
    return((int)sa_SYS_ENDIAN_MODE_LITTLE);
#endif

} 



